// Datei VarAnzParams01.c
/* ------------------------------------------------------------------------
Eine Funktion kann in C so programmiert werden, dass sie mit unterschied-
lich vielen aktuellen Parametern aufgerufen werden kann. Hier werden solche
Funktionen als VAP-Funktionen (Funktionen mit variabler Anzahl von Params)
bezeichnet.

Problem: Wie stellt man im Rumpf einer VAP-Funktion fest, mit wieviel
aktuellen Parametern sie gerade aufgerufen wurde und von welchem Typ
(oder von welchen TypEN, Mehrzahl!) die einzelnen Parameter sind?
Loesung: Nur durch zusaetzliche Parameter.

Die Kopfdatei <stdarg.h> enthaelt Groessen, die das Programmieren solcher
Funktionen erleichtern (den Typ va_list und die Makros va_start, va_arg
und va_end. "va" soll an "variable number of arguments" erinnern.).
Auf die "variablen Parameter" greift man mit dem Makro va_arg und einem
Zeiger vom Typ va_list zu. Diesen Zeiger muss man mit dem Makro va_start
initialisieren und man sollte ihn am Schluss mit dem Makro va_end
"aufraeumen".

Bekannte Beispiele fuer VAP-Funktionen: printf, scanf (mit unterschied-
lich vielen Parametern von unterschiedlichen Typen). Die hier gezeigten
Beispiele sind einfacher: Alle "variablen Parameter" muessen vom selben
Typ sein.
------------------------------------------------------------------------ */
#include <stdarg.h> // Ermoeglicht variable Anzahl von Parametern
#include <stdio.h>  // E/A in Bytestroeme, printf, sprintf, ...
#include <limits.h> // INT_MIN, INT_MAX, LONG_MIN, LONG_MAX, ...
// ------------------------------------------------------------------------
// Eine VAP-Funktion muss mindestens einen benannten ("normalen") Parameter
// haben (der als 2. Parameter fuer das Makro va_start gebraucht wird).
long add(unsigned char anzahl, long summand1, long summand2, ...) {
   // Diese Funktion muss mit mindestens 3 Parametern aufgerufen werden.
   // Alle ausser dem ersten Parameter werden als Werte des Typs long
   // interpretiert. Der erste Parameter sollte die Anzahl der weiteren
   // Parameter enthalten (mindestens den Wert 2).
   long    erg = summand1 + summand2;
   va_list summandN;                     // Typ va_list

   // va_start initialisieren. Als zweiten Parameter muss man
   // den letzten benannten ("normalen") Parameter angeben.
   va_start(summandN, summand2);         // Makro va_start
   for (anzahl-=2; anzahl>0; anzahl--) {
      erg += va_arg(summandN, long);     // Makro va_arg
   }
   va_end(summandN);                     // Makro va_end

   return erg;
} // add
// ------------------------------------------------------------------------
int max(unsigned char anzahl, ...) {
   // anzahl sollte die Anzahl der weiteren Parameter enthalten (mind. 0).
   // Diese weiteren Parameter werden als Werte des Typs int interpretiert.
   int max = INT_MIN;
   int par;                              // Wert eines Parameters

   va_list parN;                         // Zeiger auf einen Parameter

   va_start(parN, anzahl);               // parN zeigt auf ersten int-Param
   for (; anzahl>0; anzahl--) {

      // Der Wert des aktuellen int-Parameters wird nach par gebracht und
      // parN zeigt danach auf den naechsten int-Parameter:
      par = va_arg(parN, int);
      if (max < par) max = par;          // Maximum berechnen
   }
   va_end(parN);                         // Den Zeiger parN "aufraeumen"

   return max;
} // max
// ------------------------------------------------------------------------
int main() {
   unsigned long umax = UINT_MAX;
   double        d01  = 2.5;

   printf("A VarAnzParams01: Jetzt geht es los!\n");
   // Richtig funktionierende Aufrufe der Funktion add:
   printf("----------------------------------------\n");
   printf("B add(3, 5, 2, 1):          %d\n",  add(3, 5, 2, 1));
   printf("C add(4, 5, 2, 1, 3):       %d\n",  add(4, 5, 2, 1, 3));
   printf("D add(5, 5, 2, 1, 3, 1):    %d\n",  add(5, 5, 2, 1, 3, 1));
   printf("E add(6, 3, 3, 3, 3, 3, 3): %d\n",  add(6, 3, 3, 3, 3, 3, 3));
   // Nicht richtig funktionierende Aufrufe der Funktion add:
   printf("----------------------------------------\n");
   printf("F umax:                     %lu\n", umax);
   printf("G d01:                      %e\n",  d01);
   printf("H add(3, 0, 0, umax):       %d\n",  add(3, 0, 0, umax));
   printf("I add(3, 0, 0, d01):        %d\n",  add(3, 0, 0, d01));
   printf("J add(4, 0, 0, d01):        %d\n",  add(4, 0, 0, d01));
   // Richtig funktionierende Aufrufe der Funktion max:
   printf("----------------------------------------\n");
   printf("K max(3, 12, 37, 25):       %d\n",   max(3, 12, 37, 25));
   printf("L max(3, 37, 12, 25):       %d\n",   max(3, 37, 12, 25));
   printf("M max(3, 25, 12, 37):       %d\n",   max(3, 25, 12, 37));
   printf("N max(3, -5, -2, -7):       %d\n",   max(3, -5, -2, -7));
   printf("O max(1, 37):               %d\n",   max(1, 37));
   printf("P max(0):                   %d\n",   max(0));
   // Nicht richtig funktionierende Aufrufe der Funktion max:
   printf("----------------------------------------\n");
   printf("Q max(1, umax):             %d\n",   max(1, umax));
   printf("R max(1, d01):              %d\n",   max(1, d01));
   printf("S max(2, d01):              %d\n",   max(2, d01));
   printf("----------------------------------------\n");

   return 0;
} // main
/* ------------------------------------------------------------------------
Ausgabe des Programms VarAnzParams01:

A VarAnzParams01: Jetzt geht es los!
----------------------------------------
B add(3, 5, 2, 1):          8
C add(4, 5, 2, 1, 3):       11
D add(5, 5, 2, 1, 3, 1):    12
E add(6, 3, 3, 3, 3, 3, 3): 18
----------------------------------------
F umax:                     4294967295
G d01:                      2.500000e+00
H add(3, 0, 0, umax):       -1
I add(4, 0, 0, d01):        1074003968
----------------------------------------
J max(3, 12, 37, 25):       37
K max(3, 37, 12, 25):       37
L max(3, 25, 12, 37):       37
M max(3, -5, -2, -7):       -2
N max(1, 37):               37
O max(0):                   -2147483648
----------------------------------------
P max(1, umax):             -1
Q max(1, d01):              0
R max(2, d01):              1074003968
----------------------------------------
------------------------------------------------------------------------ */
